from typing import Optional,Any

class FunctionParameter:
    def __init__(self, name: str, default: Any, arg_type: type):
        self.name = name
        self.default = default
        self.arg_type = arg_type
    
    def to_dict(self):
        return {
            "name": self.name,
            "default": self.default,
            "arg_type": self.arg_type.__name__
        }



from flask import Flask, render_template_string, jsonify, request
from typing import Callable,Any
from inspect import getdoc,getmembers,ismethod
from .FunctionDefinition import make_function_definition
from .FunctionEntrypoint import make_function_entrypoint
import os.path

template_dir = os.path.join(os.path.dirname(__file__),"templates")

def render_agwui_template(template_name,**context):
    filename = os.path.join(template_dir,template_name)
    return render_template_string(
        open(filename).read(),
        **context
    )

def inject_function_app(app: Flask,function: Callable[...,Any]):
    function_definition = make_function_definition(function)
    function_entrypoint = make_function_entrypoint(function_definition)

    def get():
        return render_agwui_template(
            "function.html",
            action_path=function_entrypoint.path,
            function_definition=function_definition.to_dict()
        )
    get.__name__ = "%s_get" % function_definition.name
    app.route(function_entrypoint.path, methods=["GET"])(get)

    def post():
        values = [
            (parameter.arg_type)(request.form[parameter.name])
            for parameter
            in function_definition.parameters
        ]
        return str(function(*values))
    post.__name__ = "%s_post" % function_definition.name
    app.route(function_entrypoint.path,methods=["POST"])(post)
    
    def get_schema():
        return jsonify(
            function_definition.to_dict()
        )
    get_schema.__name__ = "%s_get_schema" % function_definition.name
    app.route(function_entrypoint.schema_path)(get_schema)


def inject_class_app(app: Flask,obj: Any):
    name = obj.__class__.__name__
    description = getdoc(obj)

    functions = [function for (name,function) in getmembers(obj,predicate=ismethod) if not name.startswith("__")]
    function_entrypoints = [
        make_function_entrypoint(make_function_definition(function))
        for function
        in functions
    ]

    @app.route("/")
    def index():
        return render_agwui_template(
            "class.html",
            name=name,
            description=description,
            function_entrypoints = function_entrypoints
        )

    for function in functions:
        inject_function_app(app,function)


from .FunctionDefinition import FunctionDefinition

class FucntionEntrypoint:
    def __init__(self, path: str, schema_path: str, function_definition: FunctionDefinition):
        self.path = path
        self.schema_path = schema_path
        self.function_definition = function_definition

def make_function_entrypoint(function_definition: FunctionDefinition) -> FucntionEntrypoint:
    path = "/" + function_definition.name
    schema_path = path + "/schema"
    return FucntionEntrypoint(
        path,
        schema_path,
        function_definition
    )


from typing import List,Optional,Callable,Any
from .FunctionParameter import FunctionParameter
from inspect import signature,getdoc,Parameter

class FunctionDefinition:
    def __init__(self, name: str, description: Optional[str], return_type: type, parameters: List[FunctionParameter]):
        self.name = name
        self.description = description
        self.return_type = return_type
        self.parameters = parameters

    def to_dict(self):
        return {
            "name": self.name,
            "description": self.description,
            "return_type": self.return_type.__name__,
            "parameters": list(map(lambda x:x.to_dict(),self.parameters))
        }

def make_function_definition(function: Callable[...,Any]) -> FunctionDefinition:
    name = function.__name__
    description = getdoc(function)
    function_signature = signature(function)
    return_type = function_signature.return_annotation

    return FunctionDefinition(
        name,
        description,
        return_type,
        [
            FunctionParameter(name, parameter.default if parameter.default != Parameter.empty else None, parameter.annotation)
            for (name,parameter)
            in function_signature.parameters.items()   
        ]
    )


from __future__ import absolute_import
from __future__ import unicode_literals
import os

from setuptools import setup, find_packages
import agwui

try:
    with open('README.md') as f:
        readme = f.read()
except IOError:
    readme = ''

def _requires_from_file(filename):
    return open(filename).read().splitlines()

# version
here = os.path.dirname(os.path.abspath(__file__))
version = agwui.__version__

setup(
    name="agwui",
    version=version,
    packages=['agwui'],
    package_dir={'agwui': 'agwui'},
    package_data={'agwui': ['templates/*.html']},
    url='https://github.com/kotauchisunsun/agwui',
    author='kotauchisunsun',
    author_email='kotauchisunshine@hotmail.co.jp',
    maintainer='kotauchisunsun',
    maintainer_email='kotauchisunshine@hotmail.co.jp',
    description='agwui is a library for Automatically Generating Web User Interface for any Python object.',
    long_description=readme,
    install_requires=_requires_from_file('requirements.txt'),
    license="MIT",
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
    ]
)